import numpy as np
import pandas as pd
from copy import deepcopy
from sklearn.base import BaseEstimator, RegressorMixin
from ..utils import is_multitask_estimator

class MultiTaskRegressor(BaseEstimator, RegressorMixin):
    """
    A class for multi-task regression

    Parameters
    ----------
    regr: object
        A regressor object

    Attributes
    ----------
    objs: list
        A list containing the fitted regressor objects

    """
    def __init__(self, regr):
        assert is_multitask_estimator(regr) == False, "The regressor is already a multi-task regressor"
        self.regr = regr
        self.objs = []

    def fit(self, X, y):
        """
        Fit the regressor

        Parameters
        ----------
        X: array-like
            The input data
        y: array-like
            The target values

        """
        n_tasks = y.shape[1]   
        assert n_tasks > 1, "The number of columns in y must be greater than 1"
        self.n_outputs_ = n_tasks     
        try:  
            for i in range(n_tasks):
                self.regr.fit(X, y.iloc[:, i].values) 
                self.objs.append(deepcopy(self.regr))
        except Exception:
            for i in range(n_tasks):
                self.regr.fit(X, y[:, i])
                self.objs.append(deepcopy(self.regr))
        return self

    def predict(self, X):
        """
        Predict the target values

        Parameters
        ----------
        X: array-like
            The input data

        Returns
        -------
        y_pred: array-like
            The predicted target values

        """
        assert len(self.objs) > 0, "The regressor has not been fitted yet"
        y_pred = np.zeros((X.shape[0], self.n_outputs_))
        for i in range(self.n_outputs_):
            y_pred[:, i] = self.objs[i].predict(X)
        return y_pred


#!/usr/bin/env python

"""
Nonconformity functions.
"""

# Authors: Henrik Linusson
# Yaniv Romano modified RegressorNc class to include CQR

from __future__ import division

import abc
import numpy as np
import sklearn.base
from .base import ClassifierAdapter, RegressorAdapter
from .base import OobClassifierAdapter, OobRegressorAdapter

# -----------------------------------------------------------------------------
# Error functions
# -----------------------------------------------------------------------------


class ClassificationErrFunc(object):
    """Base class for classification model error functions."""

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        super(ClassificationErrFunc, self).__init__()

    @abc.abstractmethod
    def apply(self, prediction, y):
        """Apply the nonconformity function.

        Parameters
        ----------
        prediction : numpy array of shape [n_samples, n_classes]
            Class probability estimates for each sample.

        y : numpy array of shape [n_samples]
            True output labels of each sample.

        Returns
        -------
        nc : numpy array of shape [n_samples]
            Nonconformity scores of the samples.
        """
        pass


class RegressionErrFunc(object):
    """Base class for regression model error functions."""

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        super(RegressionErrFunc, self).__init__()

    @abc.abstractmethod
    def apply(self, prediction, y):  # , norm=None, beta=0):
        """Apply the nonconformity function.

        Parameters
        ----------
        prediction : numpy array of shape [n_samples, n_classes]
            Class probability estimates for each sample.

        y : numpy array of shape [n_samples]
            True output labels of each sample.

        Returns
        -------
        nc : numpy array of shape [n_samples]
            Nonconformity scores of the samples.
        """
        pass

    @abc.abstractmethod
    def apply_inverse(self, nc, significance):  # , norm=None, beta=0):
        """Apply the inverse of the nonconformity function (i.e.,
        calculate prediction interval).

        Parameters
        ----------
        nc : numpy array of shape [n_calibration_samples]
            Nonconformity scores obtained for conformal predictor.

        significance : float
            Significance level (0, 1).

        Returns
        -------
        interval : numpy array of shape [n_samples, 2]
            Minimum and maximum interval boundaries for each prediction.
        """
        pass


class InverseProbabilityErrFunc(ClassificationErrFunc):
    """Calculates the probability of not predicting the correct class.

    For each correct output in ``y``, nonconformity is defined as

    .. math::
        1 - \hat{P}(y_i | x) \, .
    """

    def __init__(self):
        super(InverseProbabilityErrFunc, self).__init__()

    def apply(self, prediction, y):
        prob = np.zeros(y.size, dtype=np.float32)
        for i, y_ in enumerate(y):
            if y_ >= prediction.shape[1]:
                prob[i] = 0
            else:
                prob[i] = prediction[i, int(y_)]
        return 1 - prob


class MarginErrFunc(ClassificationErrFunc):
    """
    Calculates the margin error.

    For each correct output in ``y``, nonconformity is defined as

    .. math::
        0.5 - \dfrac{\hat{P}(y_i | x) - max_{y \, != \, y_i} \hat{P}(y | x)}{2}
    """

    def __init__(self):
        super(MarginErrFunc, self).__init__()

    def apply(self, prediction, y):
        prob = np.zeros(y.size, dtype=np.float32)
        for i, y_ in enumerate(y):
            if y_ >= prediction.shape[1]:
                prob[i] = 0
            else:
                prob[i] = prediction[i, int(y_)]
                prediction[i, int(y_)] = -np.inf
        return 0.5 - ((prob - prediction.max(axis=1)) / 2)


class AbsErrorErrFunc(RegressionErrFunc):
    """Calculates absolute error nonconformity for regression problems.

    For each correct output in ``y``, nonconformity is defined as

    .. math::
        | y_i - \hat{y}_i |
    """

    def __init__(self):
        super(AbsErrorErrFunc, self).__init__()

    def apply(self, prediction, y):
        return np.abs(prediction - y)

    def apply_inverse(self, nc, significance):
        nc = np.sort(nc)[::-1]
        border = int(np.floor(significance * (nc.size + 1))) - 1
        # TODO: should probably warn against too few calibration examples
        border = min(max(border, 0), nc.size - 1)
        return np.vstack([nc[border], nc[border]])


class SignErrorErrFunc(RegressionErrFunc):
    """Calculates signed error nonconformity for regression problems.

    For each correct output in ``y``, nonconformity is defined as

    .. math::
        y_i - \hat{y}_i

    References
    ----------
    .. [1] Linusson, Henrik, Ulf Johansson, and Tuve Lofstrom.
        Signed-error conformal regression. Pacific-Asia Conference on Knowledge
        Discovery and Data Mining. Springer International Publishing, 2014.
    """

    def __init__(self):
        super(SignErrorErrFunc, self).__init__()

    def apply(self, prediction, y):
        return prediction - y

    def apply_inverse(self, nc, significance):

        err_high = -nc
        err_low = nc

        err_high = np.reshape(err_high, (nc.shape[0], 1))
        err_low = np.reshape(err_low, (nc.shape[0], 1))

        nc = np.concatenate((err_low, err_high), 1)

        nc = np.sort(nc, 0)
        index = int(np.ceil((1 - significance / 2) * (nc.shape[0] + 1))) - 1
        index = min(max(index, 0), nc.shape[0] - 1)
        return np.vstack([nc[index, 0], nc[index, 1]])


# CQR symmetric error function
class QuantileRegErrFunc(RegressionErrFunc):
    """Calculates conformalized quantile regression error.

    For each correct output in ``y``, nonconformity is defined as

    .. math::
        max{\hat{q}_low - y, y - \hat{q}_high}

    """

    def __init__(self):
        super(QuantileRegErrFunc, self).__init__()

    def apply(self, prediction, y):
        y_lower = prediction[:, 0]
        y_upper = prediction[:, -1]
        error_low = y_lower - y
        error_high = y - y_upper
        err = np.maximum(error_high, error_low)
        return err

    def apply_inverse(self, nc, significance):
        nc = np.sort(nc, 0)
        index = int(np.ceil((1 - significance) * (nc.shape[0] + 1))) - 1
        index = min(max(index, 0), nc.shape[0] - 1)
        return np.vstack([nc[index], nc[index]])


# CQR asymmetric error function
class QuantileRegAsymmetricErrFunc(RegressionErrFunc):
    """Calculates conformalized quantile regression asymmetric error function.

    For each correct output in ``y``, nonconformity is defined as

    .. math::
        E_low = \hat{q}_low - y
        E_high = y - \hat{q}_high

    """

    def __init__(self):
        super(QuantileRegAsymmetricErrFunc, self).__init__()

    def apply(self, prediction, y):
        y_lower = prediction[:, 0]
        y_upper = prediction[:, -1]

        error_high = y - y_upper
        error_low = y_lower - y

        err_high = np.reshape(error_high, (y_upper.shape[0], 1))
        err_low = np.reshape(error_low, (y_lower.shape[0], 1))

        return np.concatenate((err_low, err_high), 1)

    def apply_inverse(self, nc, significance):
        nc = np.sort(nc, 0)
        index = int(np.ceil((1 - significance / 2) * (nc.shape[0] + 1))) - 1
        index = min(max(index, 0), nc.shape[0] - 1)
        return np.vstack([nc[index, 0], nc[index, 1]])


# -----------------------------------------------------------------------------
# Base nonconformity scorer
# -----------------------------------------------------------------------------
class BaseScorer(sklearn.base.BaseEstimator):
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        super(BaseScorer, self).__init__()

    @abc.abstractmethod
    def fit(self, x, y):
        pass

    @abc.abstractmethod
    def score(self, x, y=None):
        pass


class RegressorNormalizer(BaseScorer):
    def __init__(self, base_model, normalizer_model, err_func):
        super(RegressorNormalizer, self).__init__()
        self.base_model = base_model
        self.normalizer_model = normalizer_model
        self.err_func = err_func

    def fit(self, x, y):
        residual_prediction = self.base_model.predict(x)
        residual_error = np.abs(self.err_func.apply(residual_prediction, y))

        ######################################################################
        # Optional: use logarithmic function as in the original implementation
        # available in https://github.com/donlnz/nonconformist
        #
        # CODE:
        # residual_error += 0.00001 # Add small term to avoid log(0)
        # log_err = np.log(residual_error)
        ######################################################################

        log_err = residual_error
        self.normalizer_model.fit(x, log_err)

    def score(self, x, y=None):

        ######################################################################
        # Optional: use logarithmic function as in the original implementation
        # available in https://github.com/donlnz/nonconformist
        #
        # CODE:
        # norm = np.exp(self.normalizer_model.predict(x))
        ######################################################################

        norm = np.abs(self.normalizer_model.predict(x))
        return norm


class NcFactory(object):
    @staticmethod
    def create_nc(model, err_func=None, normalizer_model=None, oob=False):
        if normalizer_model is not None:
            normalizer_adapter = RegressorAdapter(normalizer_model)
        else:
            normalizer_adapter = None

        if isinstance(model, sklearn.base.ClassifierMixin):
            err_func = MarginErrFunc() if err_func is None else err_func
            if oob:
                c = sklearn.base.clone(model)
                c.fit([[0], [1]], [0, 1])
                if hasattr(c, "oob_decision_function_"):
                    adapter = OobClassifierAdapter(model)
                else:
                    raise AttributeError(
                        "Cannot use out-of-bag "
                        "calibration with {}".format(model.__class__.__name__)
                    )
            else:
                adapter = ClassifierAdapter(model)

            if normalizer_adapter is not None:
                normalizer = RegressorNormalizer(
                    adapter, normalizer_adapter, err_func
                )
                return ClassifierNc(adapter, err_func, normalizer)
            else:
                return ClassifierNc(adapter, err_func)

        elif isinstance(model, sklearn.base.RegressorMixin):
            err_func = AbsErrorErrFunc() if err_func is None else err_func
            if oob:
                c = sklearn.base.clone(model)
                c.fit([[0], [1]], [0, 1])
                if hasattr(c, "oob_prediction_"):
                    adapter = OobRegressorAdapter(model)
                else:
                    raise AttributeError(
                        "Cannot use out-of-bag "
                        "calibration with {}".format(model.__class__.__name__)
                    )
            else:
                adapter = RegressorAdapter(model)

            if normalizer_adapter is not None:
                normalizer = RegressorNormalizer(
                    adapter, normalizer_adapter, err_func
                )
                return RegressorNc(adapter, err_func, normalizer)
            else:
                return RegressorNc(adapter, err_func)


class BaseModelNc(BaseScorer):
    """Base class for nonconformity scorers based on an underlying model.

    Parameters
    ----------
    model : ClassifierAdapter or RegressorAdapter
        Underlying classification model used for calculating nonconformity
        scores.

    err_func : ClassificationErrFunc or RegressionErrFunc
        Error function object.

    normalizer : BaseScorer
        Normalization model.

    beta : float
        Normalization smoothing parameter. As the beta-value increases,
        the normalized nonconformity function approaches a non-normalized
        equivalent.
    """

    def __init__(self, model, err_func, normalizer=None, beta=1e-6):
        super(BaseModelNc, self).__init__()
        self.err_func = err_func
        self.model = model
        self.normalizer = normalizer
        self.beta = beta

        # If we use sklearn.base.clone (e.g., during cross-validation),
        # object references get jumbled, so we need to make sure that the
        # normalizer has a reference to the proper model adapter, if applicable.
        if self.normalizer is not None and hasattr(
            self.normalizer, "base_model"
        ):
            self.normalizer.base_model = self.model

        self.last_x, self.last_y = None, None
        self.last_prediction = None
        self.clean = False

    def fit(self, x, y):
        """Fits the underlying model of the nonconformity scorer.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of examples for fitting the underlying model.

        y : numpy array of shape [n_samples]
            Outputs of examples for fitting the underlying model.

        Returns
        -------
        None
        """
        self.model.fit(x, y)
        if self.normalizer is not None:
            self.normalizer.fit(x, y)
        self.clean = False

    def score(self, x, y=None):
        """Calculates the nonconformity score of a set of samples.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of examples for which to calculate a nonconformity score.

        y : numpy array of shape [n_samples]
            Outputs of examples for which to calculate a nonconformity score.

        Returns
        -------
        nc : numpy array of shape [n_samples]
            Nonconformity scores of samples.
        """
        prediction = self.model.predict(x)
        n_test = x.shape[0]
        if self.normalizer is not None:
            norm = self.normalizer.score(x) + self.beta
        else:
            norm = np.ones(n_test)
        if prediction.ndim > 1:
            ret_val = self.err_func.apply(prediction, y)
        else:
            ret_val = self.err_func.apply(prediction, y) / norm
        return ret_val


# -----------------------------------------------------------------------------
# Classification nonconformity scorers
# -----------------------------------------------------------------------------
class ClassifierNc(BaseModelNc):
    """Nonconformity scorer using an underlying class probability estimating
    model.

    Parameters
    ----------
    model : ClassifierAdapter
        Underlying classification model used for calculating nonconformity
        scores.

    err_func : ClassificationErrFunc
        Error function object.

    normalizer : BaseScorer
        Normalization model.

    beta : float
        Normalization smoothing parameter. As the beta-value increases,
        the normalized nonconformity function approaches a non-normalized
        equivalent.

    Attributes
    ----------
    model : ClassifierAdapter
        Underlying model object.

    err_func : ClassificationErrFunc
        Scorer function used to calculate nonconformity scores.

    See also
    --------
    RegressorNc, NormalizedRegressorNc
    """

    def __init__(
        self, model, err_func=MarginErrFunc(), normalizer=None, beta=1e-6
    ):
        super(ClassifierNc, self).__init__(model, err_func, normalizer, beta)


# -----------------------------------------------------------------------------
# Regression nonconformity scorers
# -----------------------------------------------------------------------------
class RegressorNc(BaseModelNc):
    """Nonconformity scorer using an underlying regression model.

    Parameters
    ----------
    model : RegressorAdapter
        Underlying regression model used for calculating nonconformity scores.

    err_func : RegressionErrFunc
        Error function object.

    normalizer : BaseScorer
        Normalization model.

    beta : float
        Normalization smoothing parameter. As the beta-value increases,
        the normalized nonconformity function approaches a non-normalized
        equivalent.

    Attributes
    ----------
    model : RegressorAdapter
        Underlying model object.

    err_func : RegressionErrFunc
        Scorer function used to calculate nonconformity scores.

    See also
    --------
    ProbEstClassifierNc, NormalizedRegressorNc
    """

    def __init__(
        self, model, err_func=AbsErrorErrFunc(), normalizer=None, beta=1e-6
    ):
        super(RegressorNc, self).__init__(model, err_func, normalizer, beta)

    def predict(self, x, nc, significance=None):
        """Constructs prediction intervals for a set of test examples.

        Predicts the output of each test pattern using the underlying model,
        and applies the (partial) inverse nonconformity function to each
        prediction, resulting in a prediction interval for each test pattern.

        Parameters
        ----------
        x : numpy array of shape [n_samples, n_features]
            Inputs of patters for which to predict output values.

        significance : float
            Significance level (maximum allowed error rate) of predictions.
            Should be a float between 0 and 1. If ``None``, then intervals for
            all significance levels (0.01, 0.02, ..., 0.99) are output in a
            3d-matrix.

        Returns
        -------
        p : numpy array of shape [n_samples, 2] or [n_samples, 2, 99]
            If significance is ``None``, then p contains the interval (minimum
            and maximum boundaries) for each test pattern, and each significance
            level (0.01, 0.02, ..., 0.99). If significance is a float between
            0 and 1, then p contains the prediction intervals (minimum and
            maximum	boundaries) for the set of test patterns at the chosen
            significance level.
        """
        n_test = x.shape[0]
        prediction = self.model.predict(x)
        if self.normalizer is not None:
            norm = self.normalizer.score(x) + self.beta
        else:
            norm = np.ones(n_test)

        if significance:
            intervals = np.zeros((x.shape[0], 2))
            err_dist = self.err_func.apply_inverse(nc, significance)
            err_dist = np.hstack([err_dist] * n_test)
            if prediction.ndim > 1:  # CQR
                intervals[:, 0] = prediction[:, 0] - err_dist[0, :]
                intervals[:, 1] = prediction[:, -1] + err_dist[1, :]
            else:  # regular conformal prediction
                err_dist *= norm
                intervals[:, 0] = prediction - err_dist[0, :]
                intervals[:, 1] = prediction + err_dist[1, :]

            return intervals
        else:  # Not tested for CQR
            significance = np.arange(0.01, 1.0, 0.01)
            intervals = np.zeros((x.shape[0], 2, significance.size))

            for i, s in enumerate(significance):
                err_dist = self.err_func.apply_inverse(nc, s)
                err_dist = np.hstack([err_dist] * n_test)
                err_dist *= norm

                intervals[:, 0, i] = prediction - err_dist[0, :]
                intervals[:, 1, i] = prediction + err_dist[0, :]

            return intervals



import numpy as np
import pickle
import platform
import warnings
from sklearn.base import BaseEstimator
from sklearn.base import RegressorMixin
from numpy.linalg import inv

try:
    from . import _lassoc as mo
except ImportError:
    import _lassoc as mo
from ..utils import get_beta, check_and_install

try:
    import jax.numpy as jnp
    from jax import device_put
    from jax.numpy.linalg import inv as jinv
except ImportError:
    pass


class LassoRegressor(BaseEstimator, RegressorMixin):
    """Lasso.

    Attributes:

        reg_lambda: float
            L1 regularization parameter.

        max_iter: int
            number of iterations of lasso shooting algorithm.

        tol: float
            tolerance for convergence of lasso shooting algorithm.

        backend: str
            type of backend; must be in ('cpu', 'gpu', 'tpu').

    """

    def __init__(self, reg_lambda=0.1, max_iter=10, tol=1e-3, backend="cpu"):
        assert backend in (
            "cpu",
            "gpu",
            "tpu",
        ), "`backend` must be in ('cpu', 'gpu', 'tpu')"

        sys_platform = platform.system()

        if (sys_platform == "Windows") and (backend in ("gpu", "tpu")):
            warnings.warn(
                "No GPU/TPU computing on Windows yet, backend set to 'cpu'"
            )
            backend = "cpu"

        self.reg_lambda = reg_lambda
        self.max_iter = max_iter
        self.tol = tol
        self.backend = backend
        if self.backend in ("gpu", "tpu"):
            check_and_install("jax")
            check_and_install("jaxlib")

    def fit(self, X, y, **kwargs):
        """Fit matrixops (classifier) to training data (X, y)

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            **kwargs: additional parameters to be passed to self.cook_training_set.

        Returns:

            self: object.

        """

        self.ym, centered_y = mo.center_response(y)
        self.xm = X.mean(axis=0)
        self.xsd = X.std(axis=0)
        self.xsd[self.xsd == 0] = 1
        X_ = (X - self.xm[None, :]) / self.xsd[None, :]
        XX = mo.crossprod(X_, backend=self.backend)
        Xy = mo.crossprod(X_, centered_y, backend=self.backend)
        XX2 = 2 * XX
        Xy2 = 2 * Xy

        if self.backend == "cpu":
            # beta0, _, _, _ = np.linalg.lstsq(X_, centered_y, rcond=None)
            beta0 = get_beta(X_, centered_y)
            if len(np.asarray(y).shape) == 1:
                res = mo.get_beta_1D(
                    beta0=np.asarray(beta0),
                    XX2=np.asarray(XX2),
                    Xy2=np.asarray(Xy2),
                    reg_lambda=self.reg_lambda,
                    max_iter=self.max_iter,
                    tol=self.tol,
                )
                self.beta = res[0]
                return self

            res = mo.get_beta_2D(
                beta0=np.asarray(beta0),
                XX2=np.asarray(XX2),
                Xy2=np.asarray(Xy2),
                reg_lambda=self.reg_lambda,
                max_iter=self.max_iter,
                tol=self.tol,
            )
            self.beta = res[0]
            return self

        invXX = jinv(XX + self.reg_lambda * jnp.eye(X_.shape[1]))
        beta0 = mo.safe_sparse_dot(invXX, Xy, backend=self.backend)
        if len(np.asarray(y).shape) == 1:
            res = mo.get_beta_1D(
                beta0=np.asarray(beta0),
                XX2=np.asarray(XX2),
                Xy2=np.asarray(Xy2),
                reg_lambda=self.reg_lambda,
                max_iter=self.max_iter,
                tol=self.tol,
            )
            self.beta = res[0]
            return self

        res = mo.get_beta_2D(
            beta0=np.asarray(beta0),
            XX2=np.asarray(XX2),
            Xy2=np.asarray(Xy2),
            reg_lambda=self.reg_lambda,
            max_iter=self.max_iter,
            tol=self.tol,
        )
        self.beta = res[0]
        return self

    def predict(self, X, **kwargs):
        """Predict test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to `predict_proba`


        Returns:

            model predictions: {array-like}

        """
        X_ = (X - self.xm[None, :]) / self.xsd[None, :]

        if self.backend == "cpu":
            if isinstance(self.ym, float):
                return self.ym + mo.safe_sparse_dot(X_, self.beta)
            return self.ym[None, :] + mo.safe_sparse_dot(X_, self.beta)

        # if self.backend in ("gpu", "tpu"):
        if isinstance(self.ym, float):
            return self.ym + mo.safe_sparse_dot(
                X_, self.beta, backend=self.backend
            )
        return self.ym[None, :] + mo.safe_sparse_dot(
            X_, self.beta, backend=self.backend
        )



from ._booster_regressor import LSBoostRegressor
from ._booster_regressor import GenericBoostingRegressor
from ._booster_classifier import LSBoostClassifier
from ._booster_classifier import GenericBoostingClassifier

__all__ = [
    "LSBoostClassifier",
    "LSBoostRegressor",
    "GenericBoostingClassifier",
    "GenericBoostingRegressor",
]



import numpy as np
from sklearn.base import BaseEstimator
from sklearn.base import ClassifierMixin

try:
    from . import _stumpc as stumpc
except ImportError:
    import _stumpc as stumpc


class StumpClassifier(BaseEstimator, ClassifierMixin):
    """Stump classifier.

    Attributes:

        bins: int
            Number of histogram bins; as in numpy.histogram.
    """

    def __init__(self, bins="auto"):
        self.bins = bins
        self.obj = None

    def fit(self, X, y, sample_weight=None, **kwargs):
        """Fit Stump to training data (X, y)

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            sample_weight: array_like, shape = [n_samples]
                Observations weights.

        Returns:

            self: object.
        """

        if sample_weight is None:
            self.obj = stumpc.fit_stump_classifier(
                X=np.asarray(X, order="C"),
                y=np.asarray(y, order="C"),
                bins=self.bins,
            )

            return self

        self.obj = stumpc.fit_stump_classifier(
            X=np.asarray(X, order="C"),
            y=np.asarray(y, order="C"),
            sample_weight=np.ravel(sample_weight, order="C"),
            bins=self.bins,
        )
        self.n_classes_ = len(np.unique(y))  # for compatibility with sklearn
        return self

    def predict(self, X, **kwargs):
        """Predict test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to `predict_proba`


        Returns:

            model predictions: {array-like}
        """

        return np.argmax(self.predict_proba(X, **kwargs), axis=1)

    def predict_proba(self, X, **kwargs):
        """Predict probabilities for test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to
                self.cook_test_set

        Returns:

            probability estimates for test data: {array-like}
        """

        return stumpc.predict_proba_stump_classifier(
            self.obj, np.asarray(X, order="C")
        )



from .misc import (
    cluster,
    merge_two_dicts,
    flatten,
    is_float,
    is_factor,
    check_and_install,
    is_multitask_estimator
)


__all__ = [
    "cluster",
    "merge_two_dicts",
    "flatten",
    "is_float",
    "is_factor",
    "check_and_install",
    "is_multitask_estimator"
]



import numpy as np
import platform
import warnings
from sklearn.base import BaseEstimator
from sklearn.base import RegressorMixin
from numpy.linalg import inv
from ..utils import get_beta, check_and_install
from ._enet import fit_elasticnet, predict_elasticnet

try:
    import jax.numpy as jnp
    from jax import device_put
    from jax.numpy.linalg import inv as jinv
except ImportError:
    pass


class ElasticNetRegressor(BaseEstimator, RegressorMixin):
    """Elasticnet.

    Attributes:

        reg_lambda: float
            regularization parameter.

        alpha: float
            compromise between L1 and L2 regularization (must be in [0, 1]),
            for `solver` == 'enet'.

        backend: str
            type of backend; must be in ('cpu', 'gpu', 'tpu')

    """

    def __init__(self, reg_lambda=0.1, alpha=0.5, backend="cpu"):
        assert backend in (
            "cpu",
            "gpu",
            "tpu",
        ), "`backend` must be in ('cpu', 'gpu', 'tpu')"

        sys_platform = platform.system()

        if (sys_platform == "Windows") and (backend in ("gpu", "tpu")):
            warnings.warn(
                "No GPU/TPU computing on Windows yet, backend set to 'cpu'"
            )
            backend = "cpu"

        self.reg_lambda = reg_lambda
        self.alpha = alpha
        self.backend = backend
        if self.backend in ("gpu", "tpu"):
            check_and_install("jax")
            check_and_install("jaxlib")

    def fit(self, X, y, **kwargs):
        """Fit matrixops (classifier) to training data (X, y)

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            y: array-like, shape = [n_samples]
                Target values.

            **kwargs: additional parameters to be passed to self.cook_training_set.

        Returns:

            self: object.

        """
        fit_result = fit_elasticnet(X, y, lam=self.reg_lambda, alpha=self.alpha)
        self.coef_ = fit_result.coef_
        self.y_train_mean = fit_result.y_train_mean
        self.scaler = fit_result.scaler
        self.converged = fit_result.converged
        return self

    def predict(self, X, **kwargs):
        """Predict test data X.

        Args:

            X: {array-like}, shape = [n_samples, n_features]
                Training vectors, where n_samples is the number
                of samples and n_features is the number of features.

            **kwargs: additional parameters to be passed to `predict_proba`

        Returns:

            model predictions: {array-like}

        """
        return predict_elasticnet(X, self)



import nnetsauce as ns
from functools import partial
from sklearn.base import RegressorMixin
from ..multitaskregressor.mtaskregr import MultiTaskRegressor
from sklearn.utils import all_estimators
from ..utils import is_multitask_estimator

removed_regressors = [
    "AdaBoostRegressor",
    "BaggingRegressor",
    "TheilSenRegressor",
    "ARDRegression",
    "ExtraTreesRegressor",
    "CCA",
    "GaussianProcessRegressor",
    "GradientBoostingRegressor",
    "HistGradientBoostingRegressor",
    "IsotonicRegression",
    "MLPRegressor",
    "KernelRidge",
   #"MultiOutputRegressor",
    #"MultiTaskElasticNet",
    "MultiTaskElasticNetCV",
    #"MultiTaskLasso",
    "MultiTaskLassoCV",
     "NuSVR",
    "OrthogonalMatchingPursuit",
    "OrthogonalMatchingPursuitCV",
    "PLSCanonical",
    "PLSRegression",
    "RadiusNeighborsRegressor",
    "RandomForestRegressor",
    "RANSACRegressor",
    "RegressorChain",
    "StackingRegressor",
    # "SVR",
    "VotingRegressor",
]


REGRESSORS = [
    ("GenericBooster(" + est[0] + ")", est[1])
    for est in all_estimators()
    if (
        issubclass(est[1], RegressorMixin)
        and (est[0] not in removed_regressors)
    )
]

MTASKREGRESSORS = [
    (
        "GenericBooster(MultiTask(" + est[0] + "))", 
        partial(MultiTaskRegressor, regr=est[1]()),
    )
    for est in all_estimators()
    if (
        issubclass(est[1], RegressorMixin)
        and (est[0] not in removed_regressors)
        and (est[0][:5] != "Multi")
        and (est[0][-2:] != "CV")
        and (est[0] not in ("SVR", "HuberRegressor", "LassoLarsIC"))
        and is_multitask_estimator(est[1]()) == False
    )
]



# # adapted from sklearn

# import sys
# import os

# from mlsauce._build_utils import cythonize_extensions


# def configuration(parent_package="", top_path=None):
#     from numpy.distutils.misc_util import Configuration
#     import numpy

#     libraries = []
#     if os.name == "posix":
#         libraries.append("m")

#     config = Configuration("mlsauce", parent_package, top_path)

#     # submodules with build utilities
#     config.add_subpackage("__check_build")
#     config.add_subpackage("_build_utils")

#     # submodules which do not have their own setup.py
#     # we must manually add sub-submodules & tests
#     # config.add_subpackage("demo")
#     # config.add_subpackage("encoders")
#     config.add_subpackage("utils")
#     config.add_subpackage("utils/memoryuse")
#     config.add_subpackage("utils/misc")
#     config.add_subpackage("utils/sampling")

#     # submodules which have their own setup.py
#     config.add_subpackage("adaopt")
#     config.add_subpackage("booster")
#     config.add_subpackage("lasso")
#     config.add_subpackage("ridge")
#     config.add_subpackage("stump")

#     # add the test directory
#     config.add_subpackage("tests")

#     # Skip cythonization as we do not want to include the generated
#     # C/C++ files in the release tarballs as they are not necessarily
#     # forward compatible with future versions of Python for instance.
#     if "sdist" not in sys.argv:
#         cythonize_extensions(top_path, config)

#     return config


# if __name__ == "__main__":
#     from numpy.distutils.core import setup

#     setup(**configuration(top_path="").todict())



#!/usr/bin/env python

"""
Evaluation of conformal predictors.
"""

# Authors: Henrik Linusson

# TODO: cross_val_score/run_experiment should possibly allow multiple to be evaluated on identical folding

from __future__ import division

from nonconformist.base import RegressorMixin, ClassifierMixin

import sys
import numpy as np
import pandas as pd

from sklearn.cross_validation import StratifiedShuffleSplit
from sklearn.cross_validation import KFold
from sklearn.cross_validation import train_test_split
from sklearn.base import clone, BaseEstimator


class BaseIcpCvHelper(BaseEstimator):
    """Base class for cross validation helpers."""

    def __init__(self, icp, calibration_portion):
        super(BaseIcpCvHelper, self).__init__()
        self.icp = icp
        self.calibration_portion = calibration_portion

    def predict(self, x, significance=None):
        return self.icp.predict(x, significance)


class ClassIcpCvHelper(BaseIcpCvHelper, ClassifierMixin):
    """Helper class for running the ``cross_val_score`` evaluation
    method on IcpClassifiers.

    See also
    --------
    IcpRegCrossValHelper

    Examples
    --------
    >>> from sklearn.datasets import load_iris
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> from nonconformist.icp import IcpClassifier
    >>> from nonconformist.nc import ClassifierNc, MarginErrFunc
    >>> from nonconformist.evaluation import ClassIcpCvHelper
    >>> from nonconformist.evaluation import class_mean_errors
    >>> from nonconformist.evaluation import cross_val_score
    >>> data = load_iris()
    >>> nc = ProbEstClassifierNc(RandomForestClassifier(), MarginErrFunc())
    >>> icp = IcpClassifier(nc)
    >>> icp_cv = ClassIcpCvHelper(icp)
    >>> cross_val_score(icp_cv,
    ...                 data.data,
    ...                 data.target,
    ...                 iterations=2,
    ...                 folds=2,
    ...                 scoring_funcs=[class_mean_errors],
    ...                 significance_levels=[0.1])
    ...     # doctest: +SKIP
       class_mean_errors  fold  iter  significance
    0           0.013333     0     0           0.1
    1           0.080000     1     0           0.1
    2           0.053333     0     1           0.1
    3           0.080000     1     1           0.1
    """

    def __init__(self, icp, calibration_portion=0.25):
        super(ClassIcpCvHelper, self).__init__(icp, calibration_portion)

    def fit(self, x, y):
        split = StratifiedShuffleSplit(
            y, n_iter=1, test_size=self.calibration_portion
        )
        for train, cal in split:
            self.icp.fit(x[train, :], y[train])
            self.icp.calibrate(x[cal, :], y[cal])


class RegIcpCvHelper(BaseIcpCvHelper, RegressorMixin):
    """Helper class for running the ``cross_val_score`` evaluation
    method on IcpRegressors.

    See also
    --------
    IcpClassCrossValHelper

    Examples
    --------
    >>> from sklearn.datasets import load_boston
    >>> from sklearn.ensemble import RandomForestRegressor
    >>> from nonconformist.icp import IcpRegressor
    >>> from nonconformist.nc import RegressorNc, AbsErrorErrFunc
    >>> from nonconformist.evaluation import RegIcpCvHelper
    >>> from nonconformist.evaluation import reg_mean_errors
    >>> from nonconformist.evaluation import cross_val_score
    >>> data = load_boston()
    >>> nc = RegressorNc(RandomForestRegressor(), AbsErrorErrFunc())
    >>> icp = IcpRegressor(nc)
    >>> icp_cv = RegIcpCvHelper(icp)
    >>> cross_val_score(icp_cv,
    ...                 data.data,
    ...                 data.target,
    ...                 iterations=2,
    ...                 folds=2,
    ...                 scoring_funcs=[reg_mean_errors],
    ...                 significance_levels=[0.1])
    ...     # doctest: +SKIP
       fold  iter  reg_mean_errors  significance
    0     0     0         0.185771           0.1
    1     1     0         0.138340           0.1
    2     0     1         0.071146           0.1
    3     1     1         0.043478           0.1
    """

    def __init__(self, icp, calibration_portion=0.25):
        super(RegIcpCvHelper, self).__init__(icp, calibration_portion)

    def fit(self, x, y):
        split = train_test_split(x, y, test_size=self.calibration_portion)
        x_tr, x_cal, y_tr, y_cal = split[0], split[1], split[2], split[3]
        self.icp.fit(x_tr, y_tr)
        self.icp.calibrate(x_cal, y_cal)


# -----------------------------------------------------------------------------
#
# -----------------------------------------------------------------------------
def cross_val_score(
    model,
    x,
    y,
    iterations=10,
    folds=10,
    fit_params=None,
    scoring_funcs=None,
    significance_levels=None,
    verbose=False,
):
    """Evaluates a conformal predictor using cross-validation.

    Parameters
    ----------
    model : object
            Conformal predictor to evaluate.

    x : numpy array of shape [n_samples, n_features]
            Inputs of data to use for evaluation.

    y : numpy array of shape [n_samples]
            Outputs of data to use for evaluation.

    iterations : int
            Number of iterations to use for evaluation. The data set is randomly
            shuffled before each iteration.

    folds : int
            Number of folds to use for evaluation.

    fit_params : dictionary
            Parameters to supply to the conformal prediction object on training.

    scoring_funcs : iterable
            List of evaluation functions to apply to the conformal predictor in each
            fold. Each evaluation function should have a signature
            ``scorer(prediction, y, significance)``.

    significance_levels : iterable
            List of significance levels at which to evaluate the conformal
            predictor.

    verbose : boolean
            Indicates whether to output progress information during evaluation.

    Returns
    -------
    scores : pandas DataFrame
            Tabulated results for each iteration, fold and evaluation function.
    """

    fit_params = fit_params if fit_params else {}
    significance_levels = (
        significance_levels
        if significance_levels is not None
        else np.arange(0.01, 1.0, 0.01)
    )

    df = pd.DataFrame()

    columns = [
        "iter",
        "fold",
        "significance",
    ] + [f.__name__ for f in scoring_funcs]
    for i in range(iterations):
        idx = np.random.permutation(y.size)
        x, y = x[idx, :], y[idx]
        cv = KFold(y.size, folds)
        for j, (train, test) in enumerate(cv):
            if verbose:
                sys.stdout.write(
                    "\riter {}/{} fold {}/{}".format(
                        i + 1, iterations, j + 1, folds
                    )
                )
            m = clone(model)
            m.fit(x[train, :], y[train], **fit_params)
            prediction = m.predict(x[test, :], significance=None)
            for k, s in enumerate(significance_levels):
                scores = [
                    scoring_func(prediction, y[test], s)
                    for scoring_func in scoring_funcs
                ]
                df_score = pd.DataFrame([[i, j, s] + scores], columns=columns)
                df = df.append(df_score, ignore_index=True)

    return df


def run_experiment(
    models,
    csv_files,
    iterations=10,
    folds=10,
    fit_params=None,
    scoring_funcs=None,
    significance_levels=None,
    normalize=False,
    verbose=False,
    header=0,
):
    """Performs a cross-validation evaluation of one or several conformal
    predictors on a	collection of data sets in csv format.

    Parameters
    ----------
    models : object or iterable
            Conformal predictor(s) to evaluate.

    csv_files : iterable
            List of file names (with absolute paths) containing csv-data, used to
            evaluate the conformal predictor.

    iterations : int
            Number of iterations to use for evaluation. The data set is randomly
            shuffled before each iteration.

    folds : int
            Number of folds to use for evaluation.

    fit_params : dictionary
            Parameters to supply to the conformal prediction object on training.

    scoring_funcs : iterable
            List of evaluation functions to apply to the conformal predictor in each
            fold. Each evaluation function should have a signature
            ``scorer(prediction, y, significance)``.

    significance_levels : iterable
            List of significance levels at which to evaluate the conformal
            predictor.

    verbose : boolean
            Indicates whether to output progress information during evaluation.

    Returns
    -------
    scores : pandas DataFrame
            Tabulated results for each data set, iteration, fold and
            evaluation function.
    """
    df = pd.DataFrame()
    if not hasattr(models, "__iter__"):
        models = [models]

    for model in models:
        is_regression = model.get_problem_type() == "regression"

        n_data_sets = len(csv_files)
        for i, csv_file in enumerate(csv_files):
            if verbose:
                print("\n{} ({} / {})".format(csv_file, i + 1, n_data_sets))
            data = pd.read_csv(csv_file, header=header)
            x, y = data.values[:, :-1], data.values[:, -1]
            x = np.array(x, dtype=np.float64)
            if normalize:
                if is_regression:
                    y = y - y.min() / (y.max() - y.min())
                else:
                    for j, y_ in enumerate(np.unique(y)):
                        y[y == y_] = j

            scores = cross_val_score(
                model,
                x,
                y,
                iterations,
                folds,
                fit_params,
                scoring_funcs,
                significance_levels,
                verbose,
            )

            ds_df = pd.DataFrame(scores)
            ds_df["model"] = model.__class__.__name__
            try:
                ds_df["data_set"] = csv_file.split("/")[-1]
            except:
                ds_df["data_set"] = csv_file

            df = df.append(ds_df)

    return df


# -----------------------------------------------------------------------------
# Validity measures
# -----------------------------------------------------------------------------
def reg_n_correct(prediction, y, significance=None):
    """Calculates the number of correct predictions made by a conformal
    regression model.
    """
    if significance is not None:
        idx = int(significance * 100 - 1)
        prediction = prediction[:, :, idx]

    low = y >= prediction[:, 0]
    high = y <= prediction[:, 1]
    correct = low * high

    return y[correct].size


def reg_mean_errors(prediction, y, significance):
    """Calculates the average error rate of a conformal regression model."""
    return 1 - reg_n_correct(prediction, y, significance) / y.size


def class_n_correct(prediction, y, significance):
    """Calculates the number of correct predictions made by a conformal
    classification model.
    """
    labels, y = np.unique(y, return_inverse=True)
    prediction = prediction > significance
    correct = np.zeros((y.size,), dtype=bool)
    for i, y_ in enumerate(y):
        correct[i] = prediction[i, int(y_)]
    return np.sum(correct)


def class_mean_errors(prediction, y, significance=None):
    """Calculates the average error rate of a conformal classification model."""
    return 1 - (class_n_correct(prediction, y, significance) / y.size)


def class_one_err(prediction, y, significance=None):
    """Calculates the error rate of conformal classifier predictions containing
    only a single output label.
    """
    labels, y = np.unique(y, return_inverse=True)
    prediction = prediction > significance
    idx = np.arange(0, y.size, 1)
    idx = filter(lambda x: np.sum(prediction[x, :]) == 1, idx)
    errors = filter(lambda x: not prediction[x, int(y[x])], idx)

    if len(idx) > 0:
        return np.size(errors) / np.size(idx)
    else:
        return 0


def class_mean_errors_one_class(prediction, y, significance, c=0):
    """Calculates the average error rate of a conformal classification model,
    considering only test examples belonging to class ``c``. Use
    ``functools.partial`` in order to test other classes.
    """
    labels, y = np.unique(y, return_inverse=True)
    prediction = prediction > significance
    idx = np.arange(0, y.size, 1)[y == c]
    errs = np.sum(1 for _ in filter(lambda x: not prediction[x, c], idx))

    if idx.size > 0:
        return errs / idx.size
    else:
        return 0


def class_one_err_one_class(prediction, y, significance, c=0):
    """Calculates the error rate of conformal classifier predictions containing
    only a single output label. Considers only test examples belonging to
    class ``c``. Use ``functools.partial`` in order to test other classes.
    """
    labels, y = np.unique(y, return_inverse=True)
    prediction = prediction > significance
    idx = np.arange(0, y.size, 1)
    idx = filter(lambda x: prediction[x, c], idx)
    idx = filter(lambda x: np.sum(prediction[x, :]) == 1, idx)
    errors = filter(lambda x: int(y[x]) != c, idx)

    if len(idx) > 0:
        return np.size(errors) / np.size(idx)
    else:
        return 0


# -----------------------------------------------------------------------------
# Efficiency measures
# -----------------------------------------------------------------------------
def _reg_interval_size(prediction, y, significance):
    idx = int(significance * 100 - 1)
    prediction = prediction[:, :, idx]

    return prediction[:, 1] - prediction[:, 0]


def reg_min_size(prediction, y, significance):
    return np.min(_reg_interval_size(prediction, y, significance))


def reg_q1_size(prediction, y, significance):
    return np.percentile(_reg_interval_size(prediction, y, significance), 25)


def reg_median_size(prediction, y, significance):
    return np.median(_reg_interval_size(prediction, y, significance))


def reg_q3_size(prediction, y, significance):
    return np.percentile(_reg_interval_size(prediction, y, significance), 75)


def reg_max_size(prediction, y, significance):
    return np.max(_reg_interval_size(prediction, y, significance))


def reg_mean_size(prediction, y, significance):
    """Calculates the average prediction interval size of a conformal
    regression model.
    """
    return np.mean(_reg_interval_size(prediction, y, significance))


def class_avg_c(prediction, y, significance):
    """Calculates the average number of classes per prediction of a conformal
    classification model.
    """
    prediction = prediction > significance
    return np.sum(prediction) / prediction.shape[0]


def class_mean_p_val(prediction, y, significance):
    """Calculates the mean of the p-values output by a conformal classification
    model.
    """
    return np.mean(prediction)


def class_one_c(prediction, y, significance):
    """Calculates the rate of singleton predictions (prediction sets containing
    only a single class label) of a conformal classification model.
    """
    prediction = prediction > significance
    n_singletons = np.sum(
        1 for _ in filter(lambda x: np.sum(x) == 1, prediction)
    )
    return n_singletons / y.size


def class_empty(prediction, y, significance):
    """Calculates the rate of singleton predictions (prediction sets containing
    only a single class label) of a conformal classification model.
    """
    prediction = prediction > significance
    n_empty = np.sum(1 for _ in filter(lambda x: np.sum(x) == 0, prediction))
    return n_empty / y.size


def n_test(prediction, y, significance):
    """Provides the number of test patters used in the evaluation."""
    return y.size



import numpy as np
import pandas as pd
import time

try:
    import xgboost as xgb
except ImportError:
    pass

from functools import partial
from sklearn.ensemble import RandomForestRegressor, GradientBoostingRegressor
from copy import deepcopy
from joblib import Parallel, delayed
from tqdm import tqdm
from sklearn.utils import all_estimators
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler, OneHotEncoder, OrdinalEncoder
from sklearn.compose import ColumnTransformer
from sklearn.base import RegressorMixin
from sklearn.metrics import (
    r2_score,
    root_mean_squared_error,
)
from .config import REGRESSORS
from ..booster import GenericBoostingRegressor

import warnings

warnings.filterwarnings("ignore")
pd.set_option("display.precision", 2)
pd.set_option("display.float_format", lambda x: "%.2f" % x)

numeric_transformer = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="mean")),
        ("scaler", StandardScaler()),
    ]
)

try:
    categorical_transformer_low = Pipeline(
        steps=[
            (
                "imputer",
                SimpleImputer(strategy="constant", fill_value="missing"),
            ),
            (
                "encoding",
                OneHotEncoder(handle_unknown="ignore", sparse_output=False),
            ),
        ]
    )
except TypeError:
    categorical_transformer_low = Pipeline(
        steps=[
            (
                "imputer",
                SimpleImputer(strategy="constant", fill_value="missing"),
            ),
            (
                "encoding",
                OneHotEncoder(handle_unknown="ignore", sparse=False),
            ),
        ]
    )

categorical_transformer_high = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="constant", fill_value="missing")),
        # 'OrdinalEncoder' Raise a ValueError when encounters an unknown value. Check https://github.com/scikit-learn/scikit-learn/pull/13423
        ("encoding", OrdinalEncoder()),
    ]
)


# Helper function


def get_card_split(df, cols, n=11):
    cond = df[cols].nunique() > n
    card_high = cols[cond]
    card_low = cols[~cond]
    return card_low, card_high


def adjusted_rsquared(r2, n, p):
    return 1 - (1 - r2) * ((n - 1) / (n - p - 1))


class LazyBoostingRegressor(RegressorMixin):
    """
        Fitting -- almost -- all the regression algorithms
        and returning their scores.

    Parameters:

        verbose: int, optional (default=0)
            Any positive number for verbosity.

        ignore_warnings: bool, optional (default=True)
            When set to True, the warning related to algorigms that are not able to run are ignored.

        custom_metric: function, optional (default=None)
            When function is provided, models are evaluated based on the custom evaluation metric provided.

        predictions: bool, optional (default=False)
            When set to True, the predictions of all the models models are returned as dataframe.

        sort_by: string, optional (default='RMSE')
            Sort models by a metric. Available options are 'R-Squared', 'Adjusted R-Squared', 'RMSE', 'Time Taken' and 'Custom Metric'.
            or a custom metric identified by its name and provided by custom_metric.

        random_state: int, optional (default=42)
            Reproducibiility seed.

        estimators: list, optional (default='all')
            list of Estimators names or just 'all' (default='all')

        preprocess: bool
            preprocessing is done when set to True

        n_jobs : int, when possible, run in parallel
            For now, only used by individual models that support it.

        n_layers: int, optional (default=3)
            Number of layers of CustomRegressors to be used.

        All the other parameters are the same as CustomRegressor's.

    Attributes:

        models_: dict-object
            Returns a dictionary with each model pipeline as value
            with key as name of models.

        best_model_: object
            Returns the best model pipeline based on the sort_by metric.

    Examples:

        ```python
        import os
        import mlsauce as ms
        from sklearn.datasets import load_diabetes
        from sklearn.model_selection import train_test_split

        data = load_diabetes()
        X = data.data
        y= data.target
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size = .2, random_state = 123)

        regr = ms.LazyBoostingRegressor(verbose=0, ignore_warnings=True,
                                        custom_metric=None, preprocess=True)
        models, predictioms = regr.fit(X_train, X_test, y_train, y_test)
        model_dictionary = regr.provide_models(X_train, X_test, y_train, y_test)
        print(models)
        ```

    """

    def __init__(
        self,
        verbose=0,
        ignore_warnings=True,
        custom_metric=None,
        predictions=False,
        sort_by="RMSE",
        random_state=42,
        estimators="all",
        preprocess=False,
        n_jobs=None,
    ):
        self.verbose = verbose
        self.ignore_warnings = ignore_warnings
        self.custom_metric = custom_metric
        self.predictions = predictions
        self.sort_by = sort_by
        self.models_ = {}
        self.best_model_ = None
        self.random_state = random_state
        self.estimators = estimators
        self.preprocess = preprocess
        self.n_jobs = n_jobs

    def fit(self, X_train, X_test, y_train, y_test, **kwargs):
        """Fit Regression algorithms to X_train and y_train, predict and score on X_test, y_test.

        Parameters:

            X_train : array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            X_test : array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

            y_train : array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            y_test : array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

            **kwargs: dict,
                Additional parameters to be passed to the GenericBoostingRegressor.

        Returns:
        -------
        scores:  Pandas DataFrame
            Returns metrics of all the models in a Pandas DataFrame.

        predictions : Pandas DataFrame
            Returns predictions of all the models in a Pandas DataFrame.

        """
        R2 = []
        ADJR2 = []
        RMSE = []
        # WIN = []
        names = []
        TIME = []
        predictions = {}

        if self.custom_metric:
            CUSTOM_METRIC = []

        if isinstance(X_train, np.ndarray):
            X_train = pd.DataFrame(X_train)
            X_test = pd.DataFrame(X_test)

        numeric_features = X_train.select_dtypes(include=[np.number]).columns
        categorical_features = X_train.select_dtypes(include=["object"]).columns

        categorical_low, categorical_high = get_card_split(
            X_train, categorical_features
        )

        if self.preprocess is True:
            preprocessor = ColumnTransformer(
                transformers=[
                    ("numeric", numeric_transformer, numeric_features),
                    (
                        "categorical_low",
                        categorical_transformer_low,
                        categorical_low,
                    ),
                    (
                        "categorical_high",
                        categorical_transformer_high,
                        categorical_high,
                    ),
                ]
            )

        # base models
        try:
            baseline_names = [
                "RandomForestRegressor",
                "XGBRegressor",
                "GradientBoostingRegressor",
            ]
            baseline_models = [
                RandomForestRegressor(),
                xgb.XGBRegressor(),
                GradientBoostingRegressor(),
            ]
        except Exception as exception:
            baseline_names = [
                "RandomForestRegressor",
                "GradientBoostingRegressor",
            ]
            baseline_models = [
                RandomForestRegressor(),
                GradientBoostingRegressor(),
            ]

        if self.verbose > 0:
            print("\n Fitting baseline models...")
        for name, model in tqdm(zip(baseline_names, baseline_models)):
            start = time.time()
            try:
                model.fit(X_train, y_train)
                self.models_[name] = model
                y_pred = model.predict(X_test)
                r_squared = r2_score(y_test, y_pred)
                adj_rsquared = adjusted_rsquared(
                    r_squared, X_test.shape[0], X_test.shape[1]
                )
                rmse = mean_squared_error(y_test, y_pred, squared=False)

                names.append(name)
                R2.append(r_squared)
                ADJR2.append(adj_rsquared)
                RMSE.append(rmse)
                TIME.append(time.time() - start)

                if self.custom_metric:
                    custom_metric = self.custom_metric(y_test, y_pred)
                    CUSTOM_METRIC.append(custom_metric)

                if self.verbose > 0:
                    scores_verbose = {
                        "Model": name,
                        "R-Squared": r_squared,
                        "Adjusted R-Squared": adj_rsquared,
                        "RMSE": rmse,
                        "Time taken": time.time() - start,
                    }

                    if self.custom_metric:
                        scores_verbose["Custom metric"] = custom_metric

                    print(scores_verbose)
                if self.predictions:
                    predictions[name] = y_pred
            except Exception as exception:
                if self.ignore_warnings is False:
                    print(name + " model failed to execute")
                    print(exception)

        if self.estimators == "all":
            self.regressors = REGRESSORS
        else:
            self.regressors = [
                ("GenericBooster(" + est[0] + ")", est[1](**kwargs))
                for est in all_estimators()
                if (
                    issubclass(est[1], RegressorMixin)
                    and (est[0] in self.estimators)
                )
            ]

        if self.preprocess is True:
            
            if self.n_jobs is None:

                for name, regr in tqdm(self.regressors):  # do parallel exec

                    start = time.time()

                    try:

                        model = GenericBoostingRegressor(
                            base_model=regr(), verbose=self.verbose, **kwargs
                        )

                        model.fit(X_train, y_train)

                        pipe = Pipeline(
                            steps=[
                                ("preprocessor", preprocessor),
                                ("regressor", model),
                            ]
                        )
                        if self.verbose > 0:
                            print("\n Fitting boosted " + name + " model...")
                        pipe.fit(X_train, y_train)

                        self.models_[name] = pipe
                        y_pred = pipe.predict(X_test)
                        r_squared = r2_score(y_test, y_pred)
                        adj_rsquared = adjusted_rsquared(
                            r_squared, X_test.shape[0], X_test.shape[1]
                        )
                        rmse = root_mean_squared_error(y_test, y_pred)

                        names.append(name)
                        R2.append(r_squared)
                        ADJR2.append(adj_rsquared)
                        RMSE.append(rmse)
                        TIME.append(time.time() - start)

                        if self.custom_metric:
                            custom_metric = self.custom_metric(y_test, y_pred)
                            CUSTOM_METRIC.append(custom_metric)

                        if self.verbose > 0:
                            scores_verbose = {
                                "Model": name,
                                "R-Squared": r_squared,
                                "Adjusted R-Squared": adj_rsquared,
                                "RMSE": rmse,
                                "Time taken": time.time() - start,
                            }

                            if self.custom_metric:
                                scores_verbose["Custom metric"] = custom_metric

                            print(scores_verbose)
                        if self.predictions:
                            predictions[name] = y_pred

                    except Exception as exception:

                        if self.ignore_warnings is False:
                            print(name + " model failed to execute")
                            print(exception)
            
            else: 

                results = Parallel(n_jobs=self.n_jobs)(delayed(self.train_model)(
                                name, model, X_train, y_train, X_test, y_test, 
                                use_preprocessing=True, preprocessor=preprocessor, **kwargs
                            ) for name, model in tqdm(self.regressors)
                        )
                R2 = [result["r_squared"] for result in results if result is not None]
                ADJR2 = [result["adj_rsquared"] for result in results if result is not None]
                RMSE = [result["rmse"] for result in results if result is not None]
                TIME = [result["time"] for result in results if result is not None]
                names = [result["name"] for result in results if result is not None]
                if self.custom_metric:
                    CUSTOM_METRIC = [
                        result["custom_metric"] for result in results if result is not None
                    ]
                if self.predictions:
                    predictions = {
                        result["name"]: result["predictions"] for result in results if result is not None
                    }



        else:  # self.preprocess is False; no preprocessing

            if self.n_jobs is None:

                for name, regr in tqdm(self.regressors):  # do parallel exec
                    start = time.time()
                    try:

                        model = GenericBoostingRegressor(
                            base_model=regr(), verbose=self.verbose, **kwargs
                        )

                        if self.verbose > 0:
                            print("\n Fitting boosted " + name + " model...")
                        model.fit(X_train, y_train)

                        self.models_[name] = model
                        y_pred = model.predict(X_test)

                        r_squared = r2_score(y_test, y_pred)
                        adj_rsquared = adjusted_rsquared(
                            r_squared, X_test.shape[0], X_test.shape[1]
                        )
                        rmse = root_mean_squared_error(y_test, y_pred)

                        names.append(name)
                        R2.append(r_squared)
                        ADJR2.append(adj_rsquared)
                        RMSE.append(rmse)
                        TIME.append(time.time() - start)

                        if self.custom_metric:
                            custom_metric = self.custom_metric(y_test, y_pred)
                            CUSTOM_METRIC.append(custom_metric)

                        if self.verbose > 0:
                            scores_verbose = {
                                "Model": name,
                                "R-Squared": r_squared,
                                "Adjusted R-Squared": adj_rsquared,
                                "RMSE": rmse,
                                "Time taken": time.time() - start,
                            }

                            if self.custom_metric:
                                scores_verbose["Custom metric"] = custom_metric

                            print(scores_verbose)
                        if self.predictions:
                            predictions[name] = y_pred
                    except Exception as exception:
                        if self.ignore_warnings is False:
                            print(name + " model failed to execute")
                            print(exception)
            
            else: 

                results = Parallel(n_jobs=self.n_jobs)(delayed(self.train_model)(
                                            name, model, X_train, y_train, X_test, y_test, 
                                            use_preprocessing=False, **kwargs
                                        ) for name, model in tqdm(self.regressors)
                                    )
                R2 = [result["r_squared"] for result in results if result is not None]
                ADJR2 = [result["adj_rsquared"] for result in results if result is not None]
                RMSE = [result["rmse"] for result in results if result is not None]
                TIME = [result["time"] for result in results if result is not None]
                names = [result["name"] for result in results if result is not None]
                if self.custom_metric:
                    CUSTOM_METRIC = [
                        result["custom_metric"] for result in results if result is not None
                    ]
                if self.predictions:
                    predictions = {
                        result["name"]: result["predictions"] for result in results if result is not None
                    }


        scores = {
            "Model": names,
            "Adjusted R-Squared": ADJR2,
            "R-Squared": R2,
            "RMSE": RMSE,
            "Time Taken": TIME,
        }

        if self.custom_metric:
            scores["Custom metric"] = CUSTOM_METRIC

        scores = pd.DataFrame(scores)
        scores = scores.sort_values(by=self.sort_by, ascending=True).set_index(
            "Model"
        )

        self.best_model_ = self.models_[scores.index[0]]

        if self.predictions:
            predictions_df = pd.DataFrame.from_dict(predictions)
        return scores, predictions_df if self.predictions is True else scores

    def get_best_model(self):
        """
        This function returns the best model pipeline based on the sort_by metric.

        Returns:

            best_model: object,
                Returns the best model pipeline based on the sort_by metric.

        """
        return self.best_model_

    def provide_models(self, X_train, X_test, y_train, y_test):
        """
        This function returns all the model objects trained in fit function.
        If fit is not called already, then we call fit and then return the models.

        Parameters:

            X_train : array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            X_test : array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

            y_train : array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            y_test : array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

        Returns:

            models: dict-object,
                Returns a dictionary with each model pipeline as value
                with key as name of models.

        """
        if len(self.models_.keys()) == 0:
            self.fit(X_train, X_test, y_train, y_test)

        return self.models_

    def train_model(self, name, regr, X_train, y_train, X_test, y_test, 
                    use_preprocessing=False, preprocessor=None, **kwargs):
        """
        Function to train a single regression model and return its results.
        """
        start = time.time()

        try:
            model = GenericBoostingRegressor(base_model=regr(), verbose=self.verbose, **kwargs)

            if use_preprocessing and preprocessor is not None:
                pipe = Pipeline(
                    steps=[
                        ("preprocessor", preprocessor),
                        ("regressor", model),
                    ]
                )
                if self.verbose > 0:
                    print("\n Fitting boosted " + name + " model with preprocessing...")
                pipe.fit(X_train, y_train)
                y_pred = pipe.predict(X_test)
                fitted_model = pipe
            else:
                # Case with no preprocessing
                if self.verbose > 0:
                    print("\n Fitting boosted " + name + " model without preprocessing...")
                model.fit(X_train, y_train)
                y_pred = model.predict(X_test)
                fitted_model = model

            r_squared = r2_score(y_test, y_pred)
            adj_rsquared = adjusted_rsquared(r_squared, X_test.shape[0], X_test.shape[1])
            rmse = root_mean_squared_error(y_test, y_pred)

            custom_metric = None
            if self.custom_metric:
                custom_metric = self.custom_metric(y_test, y_pred)

            return {
                "name": name,
                "model": fitted_model,
                "r_squared": r_squared,
                "adj_rsquared": adj_rsquared,
                "rmse": rmse,
                "custom_metric": custom_metric,
                "time": time.time() - start,
                "predictions": y_pred,
            }

        except Exception as exception:
            if self.ignore_warnings is False:
                print(name + " model failed to execute")
                print(exception)
            return None




# Authors: Thierry Moudiki
#
# License: BSD 3
import numpy as np
from sklearn.preprocessing import StandardScaler, MinMaxScaler, RobustScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture


def cluster(
    X,
    n_clusters=None,
    method="kmeans",
    type_scaling="standard",
    training=True,
    scaler=None,
    label_encoder=None,
    clusterer=None,
    seed=123,
):

    assert method in ("kmeans", "gmm"), "method must be in ('kmeans', 'gmm')"
    assert type_scaling in (
        "standard",
        "minmax",
        "robust",
    ), "type_scaling must be in ('standard', 'minmax', 'robust')"

    if training:
        assert (
            n_clusters is not None
        ), "n_clusters must be provided at training time"
        if type_scaling == "standard":
            scaler = StandardScaler()
        elif type_scaling == "minmax":
            scaler = MinMaxScaler()
        elif type_scaling == "robust":
            scaler = RobustScaler()
        else:
            raise ValueError(
                "type_scaling must be in ('standard', 'minmax', 'robust')"
            )

        scaled_X = scaler.fit_transform(X)
        label_encoder = OneHotEncoder(handle_unknown="ignore")

        if method == "kmeans":
            clusterer = KMeans(
                n_clusters=n_clusters, random_state=seed, n_init="auto"
            ).fit(scaled_X)
            res = label_encoder.fit_transform(
                clusterer.labels_.reshape(-1, 1)
            ).toarray()
        elif method == "gmm":
            clusterer = GaussianMixture(
                n_components=n_clusters, random_state=seed
            ).fit(scaled_X)
            res = label_encoder.fit_transform(
                clusterer.predict(scaled_X).reshape(-1, 1)
            ).toarray()
        else:
            raise ValueError("method must be in ('kmeans', 'gmm')")

        return res, scaler, label_encoder, clusterer

    else:  # @ inference time

        assert (
            scaler is not None
        ), "scaler must be provided at inferlabel_encodere time"
        assert (
            label_encoder is not None
        ), "label_encoder must be provided at inferlabel_encodere time"
        assert (
            clusterer is not None
        ), "clusterer must be provided at inferlabel_encodere time"
        scaled_X = scaler.transform(X)

        return label_encoder.transform(
            clusterer.predict(scaled_X).reshape(-1, 1)
        ).toarray()


# merge two dictionaries
def merge_two_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


# check if x is int
def is_int(x):
    try:
        return int(x) == x
    except:
        return False


# check if x is float
def is_float(x):
    return isinstance(x, float)


# check if the response contains only integers
def is_factor(y):
    n = len(y)
    ans = True
    idx = 0

    while idx < n:
        if is_int(y[idx]) & (is_float(y[idx]) == False):
            idx += 1
        else:
            ans = False
            break

    return ans


# flatten list of lists
def flatten(l):
    return [item for sublist in l for item in sublist]


import importlib
import subprocess
import sys


def install_package(package_name):
    """Install a package dynamically using pip."""
    subprocess.check_call(
        [sys.executable, "-m", "pip", "install", package_name]
    )

def is_multitask_estimator(estimator):
    return estimator._get_tags().get('multioutput', False)

def check_and_install(package_name):
    """Check if a package is installed; if not, install it."""
    try:
        # Check if the package is already installed by importing it
        importlib.import_module(package_name)
        print(f"'{package_name}' is already installed.")
    except ImportError:
        print(f"'{package_name}' not found. Installing...")
        install_package(package_name)
        # Retry importing the package after installation
        importlib.import_module(package_name)
        print(f"'{package_name}' has been installed successfully.")



#! /usr/bin/env python
#
# Copyright (C) 2020-2024 T. Moudiki <thierry.moudiki@gmail.com>
# License: 3-clause BSD

import subprocess 
import sys

subprocess.run(['pip', 'install', 'numpy'], check=False)
subprocess.run(['pip', 'install', 'scipy'], check=False)
subprocess.run(['pip', 'install', 'requests'], check=False)
subprocess.run(['pip', 'install', 'Cython'], check=False)

import os
import platform
import setuptools 
import shutil

from os import path
from pathlib import Path
from setuptools import Extension, find_packages, setup

try:
    import builtins
except ImportError:    
    import __builtin__ as builtins

import numpy
from Cython.Build import cythonize

builtins.__MLSAUCE_SETUP__ = True

DISTNAME = 'mlsauce'
DESCRIPTION = 'Miscellaneous Statistical/Machine Learning tools'
LONG_DESCRIPTION = 'Miscellaneous Statistical/Machine Learning tools'
MAINTAINER = 'T. Moudiki'
MAINTAINER_EMAIL = 'thierry.moudiki@gmail.com'
LICENSE = 'BSD3 Clause Clear'

__version__ = '0.22.2'

VERSION = __version__

if platform.python_implementation() == 'PyPy':
    SCIPY_MIN_VERSION = '1.1.0'
    NUMPY_MIN_VERSION = '1.14.0'
else:
    SCIPY_MIN_VERSION = '0.19.0'
    NUMPY_MIN_VERSION = '1.13.0'

JOBLIB_MIN_VERSION = '1.2.0'
SKLEARN_MIN_VERSION = '0.18.0'
THREADPOOLCTL_MIN_VERSION = '2.0.0'
PANDAS_MIN_VERSION = '0.25.3'
QUERIER_MIN_VERSION = '0.4.0'
REQUESTS_MIN_VERSION = '2.31.0'
if platform.system() in ('Linux', 'Darwin'):
    JAX_MIN_VERSION = '0.1.72'
    JAXLIB_MIN_VERSION = '0.1.51'

SETUPTOOLS_COMMANDS = {
    'develop', 'release', 'bdist_egg', 'bdist_rpm',
    'bdist_wininst', 'install_egg_info', 'build_sphinx',
    'egg_info', 'easy_install', 'upload', 'bdist_wheel',
    '--single-version-externally-managed',
}

if SETUPTOOLS_COMMANDS.intersection(sys.argv):
    extra_setuptools_args = dict(
        zip_safe=False,  # the package can run out of an .egg file
        include_package_data=True,
        extras_require={
            'alldeps': (
                'numpy >= {}'.format(NUMPY_MIN_VERSION),
                'scipy >= {}'.format(SCIPY_MIN_VERSION),
            ),
        },
    )
else:
    extra_setuptools_args = dict()


WHEELHOUSE_UPLOADER_COMMANDS = {'fetch_artifacts', 'upload_all'}
if WHEELHOUSE_UPLOADER_COMMANDS.intersection(sys.argv):
    import wheelhouse_uploader.cmd
    cmdclass.update(vars(wheelhouse_uploader.cmd))

ext_modules =[
    Extension(name="mlsauce.adaopt._adaoptc", 
              library_dirs=["mlsauce/adaopt/"],
              runtime_library_dirs=["mlsauce/adaopt/"],
              sources=["mlsauce/adaopt/_adaoptc.pyx"],
              include_dirs=[numpy.get_include()]),    
    Extension(name="mlsauce.booster._boosterc", 
              library_dirs=["mlsauce/booster/"],
              runtime_library_dirs=["mlsauce/booster/"],
              sources=["mlsauce/booster/_boosterc.pyx"], 
              include_dirs=[numpy.get_include()]),    
    Extension(name="mlsauce.lasso._lassoc", 
              library_dirs=["mlsauce/lasso/"],
              runtime_library_dirs=["mlsauce/lasso/"],
              sources=["mlsauce/lasso/_lassoc.pyx"], 
              include_dirs=[numpy.get_include()]),    
    Extension(name="mlsauce.ridge._ridgec", 
              library_dirs=["mlsauce/ridge/"],
              runtime_library_dirs=["mlsauce/ridge/"],
              sources=["mlsauce/ridge/_ridgec.pyx"], 
              include_dirs=[numpy.get_include()]),    
    Extension(name="mlsauce.stump._stumpc", 
              library_dirs=["mlsauce/stump/"],
              runtime_library_dirs=["mlsauce/stump/"],
              sources=["mlsauce/stump/_stumpc.pyx"], 
              include_dirs=[numpy.get_include()]),    
]

# Get the absolute path to the directory containing the setup script
abs_path = os.path.dirname(__file__)
rel_path = os.path.relpath(abs_path)
script_dir = rel_path
# Get absolute paths to Cython source files
adaopt_cython_file = str(script_dir + 'mlsauce/adaopt/_adaoptc.pyx')
booster_cython_file = str(script_dir + 'mlsauce/booster/_boosterc.pyx')
lasso_cython_file = str(script_dir + 'mlsauce/lasso/_lassoc.pyx')
ridge_cython_file = str(script_dir + 'mlsauce/ridge/_ridgec.pyx')
stump_cython_file = str(script_dir + 'mlsauce/stump/_stumpc.pyx')
# Update Extension definitions with absolute paths
ext_modules2 = [
    Extension(name="mlsauce.adaopt._adaoptc", 
              sources=[adaopt_cython_file],
              include_dirs=[numpy.get_include()]),
    Extension(name="mlsauce.booster._boosterc", 
              sources=[booster_cython_file],
              include_dirs=[numpy.get_include()]),
    Extension(name="mlsauce.lasso._lassoc", 
              sources=[lasso_cython_file],
              include_dirs=[numpy.get_include()]),
    Extension(name="mlsauce.ridge._ridgec", 
              sources=[ridge_cython_file],
              include_dirs=[numpy.get_include()]),
    Extension(name="mlsauce.stump._stumpc", 
              sources=[stump_cython_file],
              include_dirs=[numpy.get_include()]),                            
]


def setup_package():

    install_all_requires = [
        "numpy",
        "Cython",
        "joblib",
        "matplotlib",
        "nnetsauce",
        "pandas",
        "requests",
        "scikit-learn",
        "scipy",
        "tqdm"
    ]

    # if platform.system() in ('Linux', 'Darwin'):
    #     install_jax_requires = ['jax', 'jaxlib']  
    # else:
    #     install_jax_requires = []

    # install_requires = [item for sublist in [install_all_requires, install_jax_requires] for item in sublist]

    install_requires = install_all_requires

    try: 
        cythonize_ext_modules = cythonize(ext_modules2, annotate=True) 
    except: 
        cythonize_ext_modules = cythonize(ext_modules, annotate=True) 
        
    metadata = dict(name=DISTNAME,
                    maintainer=MAINTAINER,
                    maintainer_email=MAINTAINER_EMAIL,
                    description=DESCRIPTION,
                    license=LICENSE,
                    version=VERSION,
                    long_description=LONG_DESCRIPTION,
                    classifiers=['Development Status :: 2 - Pre-Alpha',
                                 'Intended Audience :: Developers',
                                 'License :: OSI Approved :: BSD License',
                                 'Natural Language :: English',
                                 'Programming Language :: Python :: 3',
                                  'Programming Language :: Python :: 3.5',
                                 'Programming Language :: Python :: 3.6',
                                 'Programming Language :: Python :: 3.7',
                                 'Programming Language :: Python :: 3.8',
                                 ],                
                    platforms=["linux", "macosx", "windows"],
                    python_requires=">=3.5",
                    install_requires=install_requires,
                    setup_requires=["numpy>= 1.13.0"],
                    package_data={'': ['*.pxd']},
                    packages=find_packages(),                    
                    ext_modules=cythonize_ext_modules,
                    **extra_setuptools_args)    
    
    
    setup(**metadata)

if __name__ == "__main__":
    setup_package()
    dir_path = os.path.dirname(os.path.realpath(__file__))
    folders = ['adaopt', 'booster', 'elasticnet', 'lasso', 'ridge', 'stump']
    for folder in folders: 
        filename = os.path.join(dir_path, "mlsauce", folder, 'setup2.py')                
        subprocess.run(['python3', filename, 'build_ext', '--inplace'])      


import nnetsauce as ns
import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier
from copy import deepcopy
from functools import partial
from joblib import Parallel, delayed
from tqdm import tqdm
import time

try:
    import xgboost as xgb
except ImportError:
    pass

from sklearn.utils import all_estimators
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler, OneHotEncoder, OrdinalEncoder
from sklearn.compose import ColumnTransformer
from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    roc_auc_score,
    f1_score,
)
from .config import REGRESSORS, MTASKREGRESSORS
from ..booster import GenericBoostingClassifier
from ..multitaskregressor import MultiTaskRegressor

import warnings

warnings.filterwarnings("ignore")
pd.set_option("display.precision", 2)
pd.set_option("display.float_format", lambda x: "%.2f" % x)

numeric_transformer = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="mean")),
        ("scaler", StandardScaler()),
    ]
)

try:
    categorical_transformer_low = Pipeline(
        steps=[
            (
                "imputer",
                SimpleImputer(strategy="constant", fill_value="missing"),
            ),
            (
                "encoding",
                OneHotEncoder(handle_unknown="ignore", sparse_output=False),
            ),
        ]
    )
except TypeError:
    categorical_transformer_low = Pipeline(
        steps=[
            (
                "imputer",
                SimpleImputer(strategy="constant", fill_value="missing"),
            ),
            (
                "encoding",
                OneHotEncoder(handle_unknown="ignore", sparse=False),
            ),
        ]
    )

categorical_transformer_high = Pipeline(
    steps=[
        ("imputer", SimpleImputer(strategy="constant", fill_value="missing")),
        # 'OrdinalEncoder' Raise a ValueError when encounters an unknown value. Check https://github.com/scikit-learn/scikit-learn/pull/13423
        ("encoding", OrdinalEncoder()),
    ]
)


def get_card_split(df, cols, n=11):
    cond = df[cols].nunique() > n
    card_high = cols[cond]
    card_low = cols[~cond]
    return card_low, card_high


class LazyBoostingClassifier(ClassifierMixin):
    """

    Fitting -- almost -- all the classification algorithms
    and returning their scores.

    Parameters:

        verbose: int, optional (default=0)
            Any positive number for verbosity.

        ignore_warnings: bool, optional (default=True)
            When set to True, the warning related to algorigms that are not
            able to run are ignored.

        custom_metric: function, optional (default=None)
            When function is provided, models are evaluated based on the custom
              evaluation metric provided.

        predictions: bool, optional (default=False)
            When set to True, the predictions of all the models models are
            returned as data frame.

        sort_by: string, optional (default='Accuracy')
            Sort models by a metric. Available options are 'Accuracy',
            'Balanced Accuracy', 'ROC AUC', 'F1 Score' or a custom metric
            identified by its name and provided by custom_metric.

        random_state: int, optional (default=42)
            Reproducibiility seed.

        estimators: list, optional (default='all')
            list of Estimators names or just 'all' for > 90 classifiers
            (default='all')

        preprocess: bool, preprocessing is done when set to True

        n_jobs: int, when possible, run in parallel
            For now, only used by individual models that support it.

        n_layers: int, optional (default=3)
            Number of layers of GenericBoostingClassifiers to be used.

        All the other parameters are the same as GenericBoostingClassifier's.

    Attributes:

        models_: dict-object
            Returns a dictionary with each model pipeline as value
            with key as name of models.

        best_model_: object
            Returns the best model pipeline.

    Examples

        ```python
        import os
        import mlsauce as ms
        from sklearn.datasets import load_breast_cancer, load_iris, load_wine, load_digits
        from sklearn.model_selection import train_test_split
        from time import time

        load_models = [load_breast_cancer, load_iris, load_wine]

        for model in load_models:

            data = model()
            X = data.data
            y= data.target

            X_train, X_test, y_train, y_test = train_test_split(X, y, test_size = .2, random_state = 13)

            clf = ms.LazyBoostingClassifier(verbose=1, ignore_warnings=False,
                                            custom_metric=None, preprocess=False)

            start = time()
            models, predictioms = clf.fit(X_train, X_test, y_train, y_test)
            print(f"\nElapsed: {time() - start} seconds\n")

            print(models)
        ```

    """

    def __init__(
        self,
        verbose=0,
        ignore_warnings=True,
        custom_metric=None,
        predictions=False,
        sort_by="Accuracy",
        random_state=42,
        estimators="all",
        preprocess=False,
        n_jobs=None,
    ):
        self.verbose = verbose
        self.ignore_warnings = ignore_warnings
        self.custom_metric = custom_metric
        self.predictions = predictions
        self.sort_by = sort_by
        self.models_ = {}
        self.best_model_ = None
        self.random_state = random_state
        self.estimators = estimators
        self.preprocess = preprocess
        self.n_jobs = n_jobs

    def fit(self, X_train, X_test, y_train, y_test, **kwargs):
        """Fit classifiers to X_train and y_train, predict and score on X_test,
        y_test.

        Parameters:

            X_train: array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            X_test: array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

            y_train: array-like,
                Training vectors, where rows is the number of samples
                and columns is the number of features.

            y_test: array-like,
                Testing vectors, where rows is the number of samples
                and columns is the number of features.

            **kwargs: dict,
                Additional arguments to be passed to the fit GenericBoostingClassifier.

        Returns:

            scores: Pandas DataFrame
                Returns metrics of all the models in a Pandas DataFrame.

            predictions: Pandas DataFrame
                Returns predictions of all the models in a Pandas DataFrame.
        """
        Accuracy = []
        B_Accuracy = []
        ROC_AUC = []
        F1 = []
        names = []
        TIME = []
        predictions = {}

        if self.custom_metric is not None:
            CUSTOM_METRIC = []

        if isinstance(X_train, np.ndarray):
            X_train = pd.DataFrame(X_train)
            X_test = pd.DataFrame(X_test)

        numeric_features = X_train.select_dtypes(include=[np.number]).columns
        categorical_features = X_train.select_dtypes(include=["object"]).columns

        categorical_low, categorical_high = get_card_split(
            X_train, categorical_features
        )

        if self.preprocess is True:
            preprocessor = ColumnTransformer(
                transformers=[
                    ("numeric", numeric_transformer, numeric_features),
                    (
                        "categorical_low",
                        categorical_transformer_low,
                        categorical_low,
                    ),
                    (
                        "categorical_high",
                        categorical_transformer_high,
                        categorical_high,
                    ),
                ]
            )

        # baseline models
        try:
            baseline_names = ["RandomForestClassifier", "XGBClassifier"]
            baseline_models = [RandomForestClassifier(), xgb.XGBClassifier()]
        except Exception as exception:
            baseline_names = ["RandomForestClassifier"]
            baseline_models = [RandomForestClassifier()]

        if self.verbose > 0:
            print("\n Fitting baseline models...")
        for name, model in tqdm(zip(baseline_names, baseline_models)):
            start = time.time()
            try:
                model.fit(X_train, y_train)
                self.models_[name] = model
                y_pred = model.predict(X_test)
                accuracy = accuracy_score(y_test, y_pred, normalize=True)
                b_accuracy = balanced_accuracy_score(y_test, y_pred)
                f1 = f1_score(y_test, y_pred, average="weighted")
                try:
                    roc_auc = roc_auc_score(y_test, y_pred)
                except Exception as exception:
                    roc_auc = None
                    if self.ignore_warnings is False:
                        print("ROC AUC couldn't be calculated for " + name)
                        print(exception)
                names.append(name)
                Accuracy.append(accuracy)
                B_Accuracy.append(b_accuracy)
                ROC_AUC.append(roc_auc)
                F1.append(f1)
                TIME.append(time.time() - start)
                if self.custom_metric is not None:
                    custom_metric = self.custom_metric(y_test, y_pred)
                    CUSTOM_METRIC.append(custom_metric)
                if self.verbose > 0:
                    if self.custom_metric is not None:
                        print(
                            {
                                "Model": name,
                                "Accuracy": accuracy,
                                "Balanced Accuracy": b_accuracy,
                                "ROC AUC": roc_auc,
                                "F1 Score": f1,
                                self.custom_metric.__name__: custom_metric,
                                "Time taken": time.time() - start,
                            }
                        )
                    else:
                        print(
                            {
                                "Model": name,
                                "Accuracy": accuracy,
                                "Balanced Accuracy": b_accuracy,
                                "ROC AUC": roc_auc,
                                "F1 Score": f1,
                                "Time taken": time.time() - start,
                            }
                        )
                if self.predictions:
                    predictions[name] = y_pred
            except Exception as exception:
                if self.ignore_warnings is False:
                    print(name + " model failed to execute")
                    print(exception)

        if self.estimators == "all":
            self.classifiers = REGRESSORS + MTASKREGRESSORS
        else:
            self.classifiers = [
                ("GBoostClassifier(" + est[0] + ")", est[1]())
                for est in all_estimators()
                if (
                    issubclass(est[1], RegressorMixin)
                    and (est[0] in self.estimators)
                )
            ] + [
                ("GBoostClassifier(MultiTask(" + est[0] + "))", partial(MultiTaskRegressor, regr=est[1]()))
                for est in all_estimators()
                if (
                    issubclass(est[1], RegressorMixin)
                    and (est[0] in self.estimators)
                )
            ]

        if self.preprocess is True:
                
            if self.n_jobs is None:

                for name, model in tqdm(self.classifiers):  # do parallel exec

                    other_args = (
                        {}
                    )  # use this trick for `random_state` too --> refactor
                    try:
                        if (
                            "n_jobs" in model().get_params().keys()
                            and name.find("LogisticRegression") == -1
                        ):
                            other_args["n_jobs"] = self.n_jobs
                    except Exception:
                        pass

                    start = time.time()

                    try:
                        if "random_state" in model().get_params().keys():
                            fitted_clf = GenericBoostingClassifier(
                                {**other_args, **kwargs},
                                verbose=self.verbose,
                                base_model=model(random_state=self.random_state),
                            )

                        else:
                            fitted_clf = GenericBoostingClassifier(
                                base_model=model(**kwargs),
                                verbose=self.verbose,
                            )

                        if self.verbose > 0:
                            print("\n Fitting boosted " + name + " model...")
                        fitted_clf.fit(X_train, y_train)

                        pipe = Pipeline(
                            [
                                ("preprocessor", preprocessor),
                                ("classifier", fitted_clf),
                            ]
                        )

                        if self.verbose > 0:
                            print("\n Fitting boosted " + name + " model...")
                        pipe.fit(X_train, y_train)
                        self.models_[name] = pipe
                        y_pred = pipe.predict(X_test)
                        accuracy = accuracy_score(y_test, y_pred, normalize=True)
                        b_accuracy = balanced_accuracy_score(y_test, y_pred)
                        f1 = f1_score(y_test, y_pred, average="weighted")
                        try:
                            roc_auc = roc_auc_score(y_test, y_pred)
                        except Exception as exception:
                            roc_auc = None
                            if self.ignore_warnings is False:
                                print("ROC AUC couldn't be calculated for " + name)
                                print(exception)
                        names.append(name)
                        Accuracy.append(accuracy)
                        B_Accuracy.append(b_accuracy)
                        ROC_AUC.append(roc_auc)
                        F1.append(f1)
                        TIME.append(time.time() - start)
                        if self.custom_metric is not None:
                            custom_metric = self.custom_metric(y_test, y_pred)
                            CUSTOM_METRIC.append(custom_metric)
                        if self.verbose > 0:
                            if self.custom_metric is not None:
                                print(
                                    {
                                        "Model": name,
                                        "Accuracy": accuracy,
                                        "Balanced Accuracy": b_accuracy,
                                        "ROC AUC": roc_auc,
                                        "F1 Score": f1,
                                        self.custom_metric.__name__: custom_metric,
                                        "Time taken": time.time() - start,
                                    }
                                )
                            else:
                                print(
                                    {
                                        "Model": name,
                                        "Accuracy": accuracy,
                                        "Balanced Accuracy": b_accuracy,
                                        "ROC AUC": roc_auc,
                                        "F1 Score": f1,
                                        "Time taken": time.time() - start,
                                    }
                                )
                        if self.predictions:
                            predictions[name] = y_pred
                    except Exception as exception:
                        if self.ignore_warnings is False:
                            print(name + " model failed to execute")
                            print(exception)

            else:
                
                # train_model(self, name, model, X_train, y_train, X_test, y_test, 
                #use_preprocessing=False, preprocessor=None, 
                #    **kwargs):
                results = Parallel(n_jobs=self.n_jobs)(delayed(self.train_model)(
                                name, model, X_train, y_train, X_test, y_test, 
                                use_preprocessing=True, preprocessor=preprocessor, **kwargs
                            ) for name, model in tqdm(self.classifiers)
                        )
                Accuracy = [res["accuracy"] for res in results]
                B_Accuracy = [res["balanced_accuracy"] for res in results]
                ROC_AUC = [res["roc_auc"] for res in results]
                F1 = [res["f1"] for res in results]
                names = [res["name"] for res in results]
                TIME = [res["time"] for res in results]
                if self.custom_metric is not None:
                    CUSTOM_METRIC = [res["custom_metric"] for res in results]
                if self.predictions:
                    predictions = {res["name"]: res["predictions"] for res in results}


        else:  # no preprocessing
                
            if self.n_jobs is None:

                for name, model in tqdm(self.classifiers):  # do parallel exec
                    start = time.time()
                    try:
                        if "random_state" in model().get_params().keys():
                            fitted_clf = GenericBoostingClassifier(
                                base_model=model(random_state=self.random_state),
                                verbose=self.verbose,
                                **kwargs
                            )

                        else:
                            fitted_clf = GenericBoostingClassifier(
                                base_model=model(), verbose=self.verbose, **kwargs
                            )

                        fitted_clf.fit(X_train, y_train)

                        self.models_[name] = fitted_clf
                        y_pred = fitted_clf.predict(X_test)
                        accuracy = accuracy_score(y_test, y_pred, normalize=True)
                        b_accuracy = balanced_accuracy_score(y_test, y_pred)
                        f1 = f1_score(y_test, y_pred, average="weighted")
                        try:
                            roc_auc = roc_auc_score(y_test, y_pred)
                        except Exception as exception:
                            roc_auc = None
                            if self.ignore_warnings is False:
                                print("ROC AUC couldn't be calculated for " + name)
                                print(exception)
                        names.append(name)
                        Accuracy.append(accuracy)
                        B_Accuracy.append(b_accuracy)
                        ROC_AUC.append(roc_auc)
                        F1.append(f1)
                        TIME.append(time.time() - start)
                        if self.custom_metric is not None:
                            custom_metric = self.custom_metric(y_test, y_pred)
                            CUSTOM_METRIC.append(custom_metric)
                        if self.verbose > 0:
                            if self.custom_metric is not None:
                                print(
                                    {
                                        "Model": name,
                                        "Accuracy": accuracy,
                                        "Balanced Accuracy": b_accuracy,
                                        "ROC AUC": roc_auc,
                                        "F1 Score": f1,
                                        self.custom_metric.__name__: custom_metric,
                                        "Time taken": time.time() - start,
                                    }
                                )
                            else:
                                print(
                                    {
                                        "Model": name,
                                        "Accuracy": accuracy,
                                        "Balanced Accuracy": b_accuracy,
                                        "ROC AUC": roc_auc,
                                        "F1 Score": f1,
                                        "Time taken": time.time() - start,
                                    }
                                )
                        if self.predictions:
                            predictions[name] = y_pred
                    except Exception as exception:
                        if self.ignore_warnings is False:
                            print(name + " model failed to execute")
                            print(exception)

            else: 

                results = Parallel(n_jobs=self.n_jobs)(delayed(self.train_model)(
                                            name, model, X_train, y_train, X_test, y_test, 
                                            use_preprocessing=False, **kwargs
                                        ) for name, model in tqdm(self.classifiers)
                                    )
                Accuracy = [res["accuracy"] for res in results]
                B_Accuracy = [res["balanced_accuracy"] for res in results]
                ROC_AUC = [res["roc_auc"] for res in results]
                F1 = [res["f1"] for res in results]
                names = [res["name"] for res in results]
                TIME = [res["time"] for res in results]
                if self.custom_metric is not None:
                    CUSTOM_METRIC = [res["custom_metric"] for res in results]
                if self.predictions:
                    predictions = {res["name"]: res["predictions"] for res in results}
                    

        if self.custom_metric is None:
            scores = pd.DataFrame(
                {
                    "Model": names,
                    "Accuracy": Accuracy,
                    "Balanced Accuracy": B_Accuracy,
                    "ROC AUC": ROC_AUC,
                    "F1 Score": F1,
                    "Time Taken": TIME,
                }
            )
        else:
            scores = pd.DataFrame(
                {
                    "Model": names,
                    "Accuracy": Accuracy,
                    "Balanced Accuracy": B_Accuracy,
                    "ROC AUC": ROC_AUC,
                    "F1 Score": F1,
                    "Custom metric": CUSTOM_METRIC,
                    "Time Taken": TIME,
                }
            )
        scores = scores.sort_values(by=self.sort_by, ascending=False).set_index(
            "Model"
        )

        self.best_model_ = self.models_[scores.index[0]]

        if self.predictions:
            predictions_df = pd.DataFrame.from_dict(predictions)
        return scores, predictions_df if self.predictions is True else scores

    def get_best_model(self):
        """
        This function returns the best model pipeline based on the sort_by metric.

        Returns:

            best_model: object,
                Returns the best model pipeline based on the sort_by metric.

        """
        return self.best_model_

    def provide_models(self, X_train, X_test, y_train, y_test):
        """Returns all the model objects trained. If fit hasn't been called yet,
        then it's called to return the models.

        Parameters:

        X_train: array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.

        X_test: array-like,
            Testing vectors, where rows is the number of samples
            and columns is the number of features.

        y_train: array-like,
            Training vectors, where rows is the number of samples
            and columns is the number of features.

        y_test: array-like,
            Testing vectors, where rows is the number of samples
            and columns is the number of features.

        Returns:

            models: dict-object,
                Returns a dictionary with each model's pipeline as value
                and key = name of the model.
        """
        if len(self.models_.keys()) == 0:
            self.fit(X_train, X_test, y_train, y_test)

        return self.models_


    def train_model(self, name, model, X_train, y_train, X_test, y_test, 
                    use_preprocessing=False, preprocessor=None, 
                    **kwargs):
        """
        Function to train a single model and return its results.
        """
        other_args = {}

        # Handle n_jobs parameter
        try:
            if "n_jobs" in model().get_params().keys() and "LogisticRegression" not in name:
                other_args["n_jobs"] = self.n_jobs
        except Exception:
            pass

        start = time.time()

        try:
            # Handle random_state parameter
            if "random_state" in model().get_params().keys():
                fitted_clf = GenericBoostingClassifier(
                    {**other_args, **kwargs},
                    verbose=self.verbose,
                    base_model=model(random_state=self.random_state),
                )
            else:
                fitted_clf = GenericBoostingClassifier(
                    base_model=model(**kwargs),
                    verbose=self.verbose,
                )

            if self.verbose > 0:
                print("\n Fitting boosted " + name + " model...")

            fitted_clf.fit(X_train, y_train)

            if use_preprocessing and preprocessor is not None:
                pipe = Pipeline(
                    [
                        ("preprocessor", preprocessor),
                        ("classifier", fitted_clf),
                    ]
                )
                if self.verbose > 0:
                    print("\n Fitting pipeline with preprocessing for " + name + " model...")
                pipe.fit(X_train, y_train)
                y_pred = pipe.predict(X_test)
            else:
                # Case with no preprocessing
                if self.verbose > 0:
                    print("\n Fitting model without preprocessing for " + name + " model...")
                y_pred = fitted_clf.predict(X_test)

            accuracy = accuracy_score(y_test, y_pred, normalize=True)
            b_accuracy = balanced_accuracy_score(y_test, y_pred)
            f1 = f1_score(y_test, y_pred, average="weighted")
            roc_auc = None

            try:
                roc_auc = roc_auc_score(y_test, y_pred)
            except Exception as exception:
                if self.ignore_warnings is False:
                    print("ROC AUC couldn't be calculated for " + name)
                    print(exception)

            custom_metric = None
            if self.custom_metric is not None:
                custom_metric = self.custom_metric(y_test, y_pred)

            return {
                "name": name,
                "model": fitted_clf if not use_preprocessing else pipe,
                "accuracy": accuracy,
                "balanced_accuracy": b_accuracy,
                "roc_auc": roc_auc,
                "f1": f1,
                "custom_metric": custom_metric,
                "time": time.time() - start,
                "predictions": y_pred,
            }
        except Exception as exception:
            if self.ignore_warnings is False:
                print(name + " model failed to execute")
                print(exception)
            return None